"""experiment.py

Classes and functions to perform actual experiments"""

import logging
import numpy as np
import scipy.constants as cnst

from muspinsim.celio import CelioHamiltonian
from muspinsim.constants import MU_TAU
from muspinsim.utils import get_xy
from muspinsim.mpi import mpi_controller as mpi
from muspinsim.simconfig import MuSpinConfig, ConfigSnapshot
from muspinsim.spinsys import MuonSpinSystem, SingleTerm
from muspinsim.input import MuSpinInput
from muspinsim.spinop import DensityOperator, SpinOperator
from muspinsim.hamiltonian import Hamiltonian
from muspinsim.lindbladian import Lindbladian


class ExperimentRunner:
    """A class meant to run experiments. Its main purpose as an object is to
    provide caching for any quantities that might not need to be recalculated
    between successive snapshots."""

    def __init__(self, in_file: MuSpinInput, variables: dict = None):
        """Set up an experiment as defined by a MuSpinInput object

        Prepare a set of calculations (for multiple files and averages) as
        defined by a MuSpinInput object and a set of variable values. Takes
        care of parallelism, splitting calculations across nodes etc.

        Arguments:
            in_file {MuSpinInput} -- The input file object defining the
                                    calculations we need to perform.
            variables {dict} -- The values of any variables appearing in the input
                                file
        """
        # Fix W0102:dangerous-default-value
        if variables is None:
            variables = {}

        if mpi.is_root:
            # On root, we run the evaluation that gives us the actual possible
            # values for simulation configurations. These are then broadcast
            # across all nodes, each of which runs its own slice of them, and
            # finally gathered back together
            config = MuSpinConfig(in_file.evaluate(**variables))
        else:
            config = MuSpinConfig()

        # broadcast config object without _system attribute
        attrs = list(config.__dict__.keys())
        for x in ["_system", "system"]:
            if x in attrs:
                attrs.remove(x)
        mpi.broadcast_object(config, attrs)

        # broadcast _system attribute without _terms attribute
        system = config.__dict__.get("_system", None)

        # Create default system only if none found
        if system is None:
            system = MuonSpinSystem()

        attrs = list(system.__dict__.keys())
        if "_terms" in attrs:
            attrs.remove("_terms")
        mpi.broadcast_object(system, attrs)

        # broadcast _terms attribute sequentially
        terms = system.__dict__.get("_terms", [])
        terms = mpi.broadcast_terms(terms)

        for i in terms:
            i.__setattr__("_spinsys", system)
        system.__setattr__("_terms", terms)
        config.__setattr__("_system", system)

        self._config = config
        self._system = config.system

        # Store single spin operators (only needed for dispersion
        # and non other non celio methods)
        if not self.config.celio_k:
            self._single_spinops = np.array(
                [
                    [self._system.operator({i: a}).matrix for a in "xyz"]
                    for i in range(len(self._system))
                ]
            )

        # Parameters
        self._B = np.zeros(3)
        self._p = np.array([1.0, 0, 0])
        self._T = np.inf

        # Basic Hamiltonian
        self._Hsys = self._system.hamiltonian

        # Derived quantities
        self._rho0 = None
        self._Hz = None
        self._dops = None

    @property
    def config(self):
        return self._config

    @property
    def system(self):
        return self._system

    @property
    def B(self):
        return self._B

    @B.setter
    def B(self, x):
        x = np.array(x)
        if (x != self._B).any():
            self._B = x
            self._rho0 = None
            self._Hz = None
            self._dops = None

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, x):
        x = np.array(x)
        if (x != self._p).any():
            self._p = x
            self._rho0 = None

    @property
    def T(self):
        return self._T

    @T.setter
    def T(self, x):
        if x != self._T:
            self._T = x
            self._rho0 = None
            self._dops = None

    @property
    def rho0(self):
        """Calculate a thermal density matrix in which the system is prepared

        Calculate an approximate thermal density matrix to prepare the system in,
        with the muon polarised along a given direction and every other spin in a
        thermal equilibrium decohered state.

        Returns:
            rho0 {DensityOperator} -- Density matrix at t=0
        """

        if self._rho0 is None:
            T = self._T
            muon_axis = self._p
            B = self._B

            if np.isclose(np.linalg.norm(B), 0.0) and T < np.inf:
                logging.warning(
                    "WARNING: initial density matrix is computed"
                    " with an approximation that can fail"
                    " at low fields and finite temperature"
                )

            mu_i = self._system.muon_index
            rhos = []

            for i in range(len(self._system.spins)):
                I = self._system.I(i)
                if i == mu_i:
                    r = DensityOperator.from_vectors(I, muon_axis, 0)
                else:
                    # Get the Zeeman Hamiltonian for this field
                    Hz = np.sum(
                        [
                            B[j] * SpinOperator.from_axes(I, e).matrix
                            for j, e in enumerate("xyz")
                        ],
                        axis=0,
                    )
                    evals, evecs = np.linalg.eigh(Hz.toarray())
                    E = evals * 1e6 * self._system.gamma(i)

                    if T > 0:
                        Z = np.exp(-cnst.h * E / (cnst.k * T))
                    else:
                        Z = np.where(E == np.amin(E), 1.0, 0.0)
                    if np.sum(Z) > 0:
                        Z /= np.sum(Z)
                    else:
                        Z = np.ones(len(E)) / len(E)

                    rhoI = np.sum(
                        evecs[:, None, :] * evecs[None, :, :].conj() * Z[None, None, :],
                        axis=-1,
                    )

                    r = DensityOperator(rhoI)

                rhos.append(r)

            self._rho0 = rhos[0]
            for rho in rhos[1:]:
                self._rho0 = self._rho0.kron(rho)

        return self._rho0

    @property
    def Hz(self):
        if self._Hz is None:
            # Compute Zeeman Hamiltonian contribution
            if not self._config.celio_k:
                B = self._B
                g = self._system.gammas
                Bg = B[None, :] * g[:, None]

                Hz_sp_list = (self._single_spinops * Bg).flatten().tolist()
                Hz = np.sum(Hz_sp_list)

                self._Hz = Hamiltonian(Hz, dim=self._system.dimension)
            else:
                extra_terms = []
                # Add zeeman terms only if there is a field present to avoid
                # making Celio's method unnecessarily expensive
                if not np.array_equal(self._B, [0, 0, 0]):
                    for i in range(len(self._system.spins)):
                        extra_terms.append(
                            SingleTerm(
                                self._system,
                                i,
                                self._B * self._system.gammas[i],
                                label="Zeeman",
                            )
                        )
                self._Hz = CelioHamiltonian(
                    extra_terms, self.config.celio_k, self._system
                )

        return self._Hz

    @property
    def dissipation_operators(self):
        if self._config.celio_k:
            raise NotImplementedError(
                "Dissipation is not supported when using Celio's method"
            )

        if self._dops is None:

            # Create a copy of the system
            sys = self._system.clone()

            # Clean it up of all terms
            sys.clear_terms()
            sys.clear_dissipative_terms()

            T = self._T
            # We only go by the intensity of the field
            B = np.linalg.norm(self._B)
            g = sys.gammas
            if T > 0:
                Zu = np.exp(-cnst.h * g * B * 1e6 / (cnst.k * T))
                if np.isclose(B, 0.0) and T < np.inf:
                    logging.warning(
                        "WARNING: dissipation effects are computed"
                        " with an approximation that can fail"
                        " at low fields and finite temperature"
                    )
            else:
                Zu = g * 0.0

            if B == 0:
                x, y = np.array([1.0, 0, 0]), np.array([0, 1.0, 0])
            else:
                z = self._B / B
                x, y = get_xy(z)

            def sparse_sum(sp_mat_list):
                sp_mat_list = sp_mat_list.flatten().tolist()
                return np.sum(sp_mat_list)

            self._dops = []
            for i, a in self._config.dissipation_terms.items():

                op_x = sparse_sum(self._single_spinops[i, :, None] * x[:, None])
                op_y = sparse_sum(self._single_spinops[i, :, None] * y[:, None])
                op_p = SpinOperator(op_x + 1.0j * op_y, dim=self.system.dimension)
                op_m = SpinOperator(op_x - 1.0j * op_y, dim=self.system.dimension)

                # The 1/pi factor here makes sure that the convention is that when
                # a user inputs a certain value of dissipation for a single spin,
                # that is exactly the exponential decay factor that is observed.

                self._dops.append((op_p, 1 / np.pi * a * Zu[i] / (1 + Zu[i])))
                self._dops.append((op_m, 1 / np.pi * a / (1 + Zu[i])))

        return self._dops

    @property
    def Hsys(self):
        return self._Hsys

    @property
    def Htot(self):
        # Build the total Hamiltonian
        H = self.Hsys + self.Hz

        # Do we have dissipation?
        if len(self._config.dissipation_terms) > 0:
            # Actually use a Lindbladian
            H = Lindbladian.from_hamiltonian(H, self.dissipation_operators)

        return H

    @property
    def p_operator(self):
        return self._system.muon_operator(self.p)

    def run(self):
        """Run the experiment

        Run all calculations in the configuration set, gather the results and
        return them.

        Returns:
            results (np.ndarray) -- An array of results, gathered on the root
                                    node.
        """

        for cfg in self._config[mpi.rank :: mpi.size]:
            dataslice = self.run_single(cfg)
            self._config.store_time_slice(cfg.id, dataslice)

        self._config.results = mpi.sum_data(self._config.results)

        return self._config.results

    def load_config(self, cfg_snap: ConfigSnapshot):
        """Load a configuration snapshot in this ExperimentRunner

        Load a configuration snapshot in the ExperimentRunner, assigning field,
        polarisation and temperature.

        Arguments:
            cfg_snap {ConfigSnapshot} -- A named tuple defining the values of all
                                         parameters to be used in this calculation.

        Returns:
            weight {float} -- The weight to assign to this specific simulation
        """

        # Let's gather the important stuff
        B = cfg_snap.B  # Magnetic field
        p = cfg_snap.mupol  # Muon polarization
        T = cfg_snap.T  # Temperature
        q, w = cfg_snap.orient  # Quaternion and Weight for orientation

        # Let's start by rotating things
        self.B = q.rotate(B)
        self.p = q.rotate(p)
        self.T = T

        # Magnetic field is sum of external field (rotated with angular
        # averages) and the intrinsic field that should not be rotated
        self.B += cfg_snap.intrinsic_B

        # Figure out if a speedup is suitable
        B = np.linalg.norm(self.B)
        check_result = (cnst.e * (cnst.hbar**2) * B) / (2 * cnst.m_p * cnst.k * T)

        # For now only use when exactly 0 (i.e. when T -> inf, or B = 0)
        self._T_inf_speedup = check_result == 0

        # Due to the method of computation we also require that muon is first in
        # the system so we check this is the case here, otherwise we need to
        # change the order of kronecker products when computing the sigma_mu for
        # the system and this would be slower anyway
        if self._T_inf_speedup and self._system.muon_index != 0:
            self._T_inf_speedup = False

            # Add a message to the log to notify there is a speedup available
            # if the system is reordered
            logging.info(
                "The system is suitable for a speedup if the muon is defined first."
            )
        return w

    def run_single(self, cfg_snap: ConfigSnapshot):
        """Run a muon experiment from a configuration snapshot

        Run a muon experiment using the specific parameters and time axis given in
        a configuration snapshot.

        Arguments:
            cfg_snap {ConfigSnapshot} -- A named tuple defining the values of all
                                         parameters to be used in this calculation.

        Returns:
            result {np.ndarray} -- A 1D array containing the time series of
                                   required results, or a single value.
        """

        w = self.load_config(cfg_snap)

        H = self.Htot

        if cfg_snap.y == "asymmetry":
            # Use faster more approximate method of Celio's if requested
            if isinstance(H, CelioHamiltonian) and self.config.celio_averages:
                # Ensure the system is valid for its use
                if self.T != np.inf:
                    raise ValueError(
                        "The fast version of Celio's method requires T -> inf. "
                        "Either remove the number of averages or don't use "
                        "celio."
                    )

                data = H.fast_evolve(
                    self._system.sigma_mu(self.p),
                    cfg_snap.t,
                    self.config.celio_averages,
                    True,
                )
            else:
                # Measurement operator
                S = self.p_operator

                # Use faster evolution if able to (Doesn't exist for Linbladian)
                if self._T_inf_speedup and not isinstance(H, Lindbladian):
                    other_spins = list(range(0, len(self._system.spins)))
                    other_spins.remove(self._system.muon_index)
                    other_dimension = np.prod(
                        [self._system.dimension[i] for i in other_spins]
                    )

                    data = H.fast_evolve(
                        self._system.sigma_mu(self.p), cfg_snap.t, other_dimension
                    )
                else:
                    data = H.evolve(self.rho0, cfg_snap.t, operators=[S])[:, 0]
        elif cfg_snap.y == "integral":
            # Measurement operator
            S = self.p_operator

            data = H.integrate_decaying(self.rho0, MU_TAU, operators=[S])[0] / MU_TAU

        return np.real(data) * w



from io import StringIO
import unittest
import numpy as np

from muspinsim.input.structure import CellAtom, MuonatedStructure

TEST_CELL_FILE_DATA = """%BLOCK LATTICE_CART
   14.18160000000000   0.000000000000000   0.000000000000000
   0.000000000000000   14.18160000000000   0.000000000000000
   0.000000000000000   0.000000000000000   14.18160000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
V             0.0791557133        0.0000005853        0.1708904764
V             0.9161667724        0.0000005823        0.8325662729
Si            0.4996287249        0.8331358706        0.8331663251
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC
"""


def _optional_float_close(value1, value2):
    # Compares two values that may be a float or none
    if value1 is None or value2 is None:
        return value1 == value2
    else:
        return np.isclose(value1, value2)


def _optional_array_close(value1, value2):
    # Compares two values that may be an array or none
    if value1 is None or value2 is None:
        return value1 == value2
    else:
        return np.allclose(value1, value2)


class CellAtomMatcher:
    expected: CellAtom

    def __init__(self, expected):
        self.expected = expected

    def __repr__(self):
        return repr(self.expected)

    def __eq__(self, other):
        return (
            self.expected.index == other.index
            and self.expected.symbol == other.symbol
            and _optional_array_close(self.expected.position, other.position)
            and _optional_array_close(
                self.expected.vector_from_muon, other.vector_from_muon
            )
            and _optional_float_close(
                self.expected.distance_from_muon, other.distance_from_muon
            )
        )


class TestStructure(unittest.TestCase):
    def test_parsing(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")

        self.assertEqual(len(structure._cell_atoms), 4)
        self.assertEqual(structure._muon_index, 3)
        self.assertEqual(
            CellAtomMatcher(structure._cell_atoms[2]),
            CellAtom(
                index=3,
                symbol="Si",
                isotope=1,
                position=np.array([7.08553473, 11.81519966, 11.81563156]),
                vector_from_muon=None,
                distance_from_muon=None,
            ),
        )

    def test_invalid_vector_angels(self):
        with self.assertRaises(ValueError):
            MuonatedStructure(
                StringIO(
                    """%BLOCK LATTICE_CART
   14.18160000000000   14.18160000000000   0.000000000000000
   0.000000000000000   14.18160000000000   0.000000000000000
   0.000000000000000   0.000000000000000   14.18160000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC"""
                ),
                fmt="castep-cell",
            )

    def test_no_muon_found(self):
        with self.assertRaises(ValueError):
            MuonatedStructure(
                StringIO(TEST_CELL_FILE_DATA),
                muon_symbol="Cu",
                fmt="castep-cell",
            )

    def test_multiple_muons_found(self):
        with self.assertRaises(ValueError):
            MuonatedStructure(
                StringIO(TEST_CELL_FILE_DATA),
                muon_symbol="V",
                fmt="castep-cell",
            )

    def test_compute_layer_offsets(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")

        offsets = structure._compute_layer_offsets(1)
        self.assertEqual(len(offsets), 26)
        self.assertTrue(
            np.allclose(
                offsets,
                [
                    [-14.1816, -14.1816, -14.1816],
                    [-14.1816, -14.1816, 0.0],
                    [-14.1816, -14.1816, 14.1816],
                    [-14.1816, 0.0, -14.1816],
                    [-14.1816, 0.0, 0.0],
                    [-14.1816, 0.0, 14.1816],
                    [-14.1816, 14.1816, -14.1816],
                    [-14.1816, 14.1816, 0.0],
                    [-14.1816, 14.1816, 14.1816],
                    [14.1816, -14.1816, -14.1816],
                    [14.1816, -14.1816, 0.0],
                    [14.1816, -14.1816, 14.1816],
                    [14.1816, 0.0, -14.1816],
                    [14.1816, 0.0, 0.0],
                    [14.1816, 0.0, 14.1816],
                    [14.1816, 14.1816, -14.1816],
                    [14.1816, 14.1816, 0.0],
                    [14.1816, 14.1816, 14.1816],
                    [0.0, -14.1816, -14.1816],
                    [0.0, -14.1816, 0.0],
                    [0.0, -14.1816, 14.1816],
                    [0.0, 14.1816, -14.1816],
                    [0.0, 14.1816, 0.0],
                    [0.0, 14.1816, 14.1816],
                    [0.0, 0.0, -14.1816],
                    [0.0, 0.0, 14.1816],
                ],
            )
        )

        # Check correct number for larger system
        offsets = structure._compute_layer_offsets(2)
        self.assertEqual(len(offsets), 98)

        # Check offsets are correct for a system with different cell lengths
        structure = MuonatedStructure(
            StringIO(
                """%BLOCK LATTICE_CART
   14.18160000000000   0.000000000000000   0.000000000000000
   0.000000000000000   20.00000000000000   0.000000000000000
   0.000000000000000   0.000000000000000   8.000000000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC"""
            ),
            fmt="castep-cell",
        )
        offsets = structure._compute_layer_offsets(1)
        self.assertTrue(
            np.allclose(
                offsets,
                np.array(
                    [
                        [-14.1816, -20.0, -8.0],
                        [-14.1816, -20.0, 0.0],
                        [-14.1816, -20.0, 8.0],
                        [-14.1816, 0.0, -8.0],
                        [-14.1816, 0.0, 0.0],
                        [-14.1816, 0.0, 8.0],
                        [-14.1816, 20.0, -8.0],
                        [-14.1816, 20.0, 0.0],
                        [-14.1816, 20.0, 8.0],
                        [14.1816, -20.0, -8.0],
                        [14.1816, -20.0, 0.0],
                        [14.1816, -20.0, 8.0],
                        [14.1816, 0.0, -8.0],
                        [14.1816, 0.0, 0.0],
                        [14.1816, 0.0, 8.0],
                        [14.1816, 20.0, -8.0],
                        [14.1816, 20.0, 0.0],
                        [14.1816, 20.0, 8.0],
                        [0.0, -20.0, -8.0],
                        [0.0, -20.0, 0.0],
                        [0.0, -20.0, 8.0],
                        [0.0, 20.0, -8.0],
                        [0.0, 20.0, 0.0],
                        [0.0, 20.0, 8.0],
                        [0.0, 0.0, -8.0],
                        [0.0, 0.0, 8.0],
                    ]
                ),
            )
        )

    def test_compute_layer(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        new_atoms = structure.compute_layer(1)

        self.assertEqual(len(new_atoms), 3 * 26)  # Copy everything but muon
        self.assertEqual(
            CellAtomMatcher(new_atoms[11]),
            CellAtom(
                index=3,
                symbol="Si",
                isotope=1,
                position=np.array([-7.09606527, 11.81519966, -2.36596844]),
                vector_from_muon=None,
                distance_from_muon=None,
            ),
        )

    def test_compute_layer_ignore(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        new_atoms = structure.compute_layer(1, ignored_symbols=["Si"])

        self.assertEqual(len(new_atoms), 2 * 26)  # Copy everything but muon and silicon
        self.assertEqual(
            CellAtomMatcher(new_atoms[11]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([-1.18888930e00, 8.25794568e-06, 2.59887219e01]),
                vector_from_muon=None,
                distance_from_muon=None,
            ),
        )

    def test_compute_closest(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        closest_atoms = structure.compute_closest(1)

        self.assertEqual(len(closest_atoms), 1)
        self.assertEqual(
            CellAtomMatcher(closest_atoms[0]),
            CellAtom(
                index=1,
                symbol="V",
                isotope=1,
                position=np.array([1.12255466e00, 8.30049048e-06, 2.42350038e00]),
                vector_from_muon=np.array(
                    [1.24105396e00, 1.76149654e-05, -1.24170213e00]
                ),
                distance_from_muon=1.7555737250028431,
            ),
        )

        more_closest_atoms = structure.compute_closest(6)

        self.assertEqual(len(more_closest_atoms), 6)
        self.assertEqual(CellAtomMatcher(closest_atoms[0]), more_closest_atoms[0])
        self.assertEqual(
            CellAtomMatcher(more_closest_atoms[5]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([1.29927107e01, 8.25794568e-06, -2.37447814e00]),
                vector_from_muon=np.array(
                    [-1.06291021e01, 1.76575102e-05, 3.55627639e00]
                ),
                distance_from_muon=11.208251995910084,
            ),
        )

    def test_compute_closest_max_layer(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")

        with self.assertRaises(RuntimeError):
            structure.compute_closest(6, max_layer=1)



import unittest
import numpy as np

from muspinsim.spinop import SpinOperator, DensityOperator
from muspinsim.hamiltonian import Hamiltonian
from muspinsim.lindbladian import Lindbladian


class TestLindbladian(unittest.TestCase):
    def test_creation(self):

        # Create a basic Hamiltonian
        sx = SpinOperator.from_axes()
        sz = SpinOperator.from_axes(0.5, "z")

        H = Hamiltonian(sz.matrix)

        # No dissipation
        L = Lindbladian.from_hamiltonian(H)

        self.assertEqual(L.dimension, (2, 2))
        self.assertTrue(np.all(L.matrix.toarray() == 1.0j * np.diag([0, -1, 1, 0])))

        # Dissipation
        L = Lindbladian.from_hamiltonian(H, [(sx, 0.1)])

        self.assertTrue(
            np.all(
                np.isclose(
                    L.matrix.toarray(),
                    [
                        [-0.025, 0, 0, 0.025],
                        [0, -0.025 - 1.0j, 0.025, 0],
                        [0, 0.025, -0.025 + 1.0j, 0],
                        [0.025, 0, 0, -0.025],
                    ],
                )
            )
        )

    def test_evolve(self):

        # Basic test
        sx = SpinOperator.from_axes()
        sp = SpinOperator.from_axes(0.5, "+")
        sm = SpinOperator.from_axes(0.5, "-")
        sz = SpinOperator.from_axes(0.5, "z")

        H = Hamiltonian(sz.matrix)
        L = Lindbladian.from_hamiltonian(H)
        rho0 = DensityOperator.from_vectors(0.5, [1, 0, 0], 0)
        t = np.linspace(0, 1, 100)

        evol = L.evolve(rho0, t, sx)

        self.assertTrue(np.all(np.isclose(evol[:, 0], 0.5 * np.cos(2 * np.pi * t))))

        # Same but with decay
        for g in [1.0, 2.0, 5.0, 10.0]:

            L = Lindbladian.from_hamiltonian(H, [(sx, g)])
            evol = L.evolve(rho0, t, sx)

            ap = -0.5 * np.pi * g + ((0.5 * np.pi * g) ** 2 - 4 * np.pi**2) ** 0.5
            am = -0.5 * np.pi * g - ((0.5 * np.pi * g) ** 2 - 4 * np.pi**2) ** 0.5
            A = ap * am / (am - ap)

            # Analytical solution for this case
            solx = np.real(0.5 * A * (np.exp(ap * t) / ap - np.exp(am * t) / am))

            self.assertTrue(np.all(np.isclose(evol[:, 0], solx)))

            gp = g * 1.5
            gm = g * 0.5
            L = Lindbladian.from_hamiltonian(H, [(sp, gp), (sm, gm)])

            evol = L.evolve(rho0, t, [sx, sz])

            solx = np.real(0.5 * np.cos(2 * np.pi * t) * np.exp(-2 * np.pi * g * t))
            solz = 0.25 * (1 - np.exp(-4 * np.pi * g * t))

            self.assertTrue(np.all(np.isclose(evol[:, 0], solx)))
            self.assertTrue(np.all(np.isclose(evol[:, 1], solz)))

    def test_integrate(self):

        # Basic test
        sx = SpinOperator.from_axes()
        sz = SpinOperator.from_axes(0.5, "z")

        H = Hamiltonian(sz.matrix)
        L = Lindbladian.from_hamiltonian(H)
        rho0 = DensityOperator.from_vectors(0.5, [1, 0, 0], 0)
        t = np.linspace(0, 1, 100)

        tau = 2.0
        avg = L.integrate_decaying(rho0, tau, sx)

        self.assertAlmostEqual(avg[0], 0.5 * tau / (1 + 4 * np.pi**2 * tau**2))

        # Same but with decay
        for g in [1.0, 2.0, 5.0, 10.0]:

            L = Lindbladian.from_hamiltonian(H, [(sx, g)])
            avg = L.integrate_decaying(rho0, tau, sx)

            ap = -0.5 * np.pi * g + ((0.5 * np.pi * g) ** 2 - 4 * np.pi**2) ** 0.5
            am = -0.5 * np.pi * g - ((0.5 * np.pi * g) ** 2 - 4 * np.pi**2) ** 0.5
            A = ap * am / (am - ap)

            # Analytical solution for this case
            sol = np.real(
                0.5 * A * tau * (1 / ((1 - ap * tau) * ap) - 1 / ((1 - am * tau) * am))
            )

            self.assertAlmostEqual(avg[0], sol)



import unittest
import numpy as np

from muspinsim.cpp import Celio_EvolveContrib
from muspinsim.celio import CelioHamiltonian
from muspinsim.spinop import DensityOperator
from muspinsim.spinsys import MuonSpinSystem, SingleTerm, SpinSystem


class TestCelioHamilto(unittest.TestCase):
    def test_sum(self):
        ssys = SpinSystem(["mu", "e"], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        extra_terms = [SingleTerm(ssys, 1, [0, 1, 0])]

        H2 = CelioHamiltonian(extra_terms, 10, ssys)
        H_sum = ssys.hamiltonian + H2

        self.assertEqual(H_sum._terms[1], extra_terms[0])

    def test_calc_H_contribs(self):
        ssys = SpinSystem(["mu", "F", ("e", 2)], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])
        ssys.add_bilinear_term(0, 1, [[0, 0, 1], [0, 0, 1], [0, 0, 1]])
        ssys.add_bilinear_term(0, 2, [[0, 0, -1], [0, 0, -1], [0, 0, -1]])
        H = ssys.hamiltonian

        H_contribs = H._calc_H_contribs()

        self.assertEqual(len(H_contribs), 3)

        def check_H_contrib(
            H_contrib, matrix, other_dimension, spin_order, spin_dimensions
        ):
            self.assertTrue(np.allclose(H_contrib.matrix.toarray(), matrix))
            self.assertEqual(H_contrib.other_dimension, other_dimension)
            self.assertTrue(np.allclose(H_contrib.spin_order, spin_order))
            self.assertTrue(np.allclose(H_contrib.spin_dimensions, spin_dimensions))

        check_H_contrib(
            H_contrib=H_contribs[0],
            matrix=[[0, 0.5], [0.5, 0]],
            other_dimension=6,
            spin_order=[0, 1, 2],
            spin_dimensions=[2, 2, 3],
        )

        check_H_contrib(
            H_contrib=H_contribs[1],
            matrix=[
                [0.25, 0, 0.25 - 0.25j, 0],
                [0, -0.25, 0, -0.25 + 0.25j],
                [0.25 + 0.25j, 0, -0.25, 0],
                [0, -0.25 - 0.25j, 0, 0.25],
            ],
            other_dimension=3,
            spin_order=[0, 1, 2],
            spin_dimensions=[2, 2, 3],
        )

        check_H_contrib(
            H_contrib=H_contribs[2],
            matrix=[
                [-0.5, 0, 0, -0.5 + 0.5j, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0.5, 0, 0, 0.5 - 0.5j],
                [-0.5 - 0.5j, 0, 0, 0.5, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0.5 + 0.5j, 0, 0, -0.5],
            ],
            other_dimension=2,
            spin_order=[0, 2, 1],
            spin_dimensions=[2, 3, 2],
        )

    def test_calc_trotter_evol_op(self):
        ssys = SpinSystem(["mu", "e"], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])
        H = ssys.hamiltonian

        evol_op_contribs = H._calc_trotter_evol_op_contribs(1, False)

        self.assertEqual(len(evol_op_contribs), 1)

        self.assertTrue(
            np.allclose(
                evol_op_contribs[0].toarray(),
                [
                    [0.95105652, 0, -0.30901699j, 0],
                    [0, 0.95105652, 0, -0.30901699j],
                    [-0.30901699j, 0, 0.95105652, 0],
                    [0, -0.30901699j, 0, 0.95105652],
                ],
            )
        )

        # Should be 2x2 for the fast variant
        ssys = SpinSystem(["mu", "e"], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])
        H = ssys.hamiltonian

        evol_op_contribs = H._calc_trotter_evol_op_contribs(1, True)

        self.assertEqual(len(evol_op_contribs), 1)
        self.assertTrue(isinstance(evol_op_contribs[0], Celio_EvolveContrib))

        self.assertTrue(
            np.allclose(
                evol_op_contribs[0].matrix,
                [
                    [0.95105652, -0.30901699j],
                    [-0.30901699j, 0.95105652],
                ],
            )
        )

        self.assertEqual(evol_op_contribs[0].other_dim, 2)
        self.assertTrue(np.allclose(evol_op_contribs[0].indices, [0, 1, 2, 3]))

    def test_evolve(self):
        ssys = SpinSystem(["e"], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        H = ssys.hamiltonian
        rho0 = DensityOperator.from_vectors()  # Start along z
        t = np.linspace(0, 1, 100)

        self.assertTrue(isinstance(H, CelioHamiltonian))

        evol = H.evolve(rho0, t, ssys.operator({0: "z"}))

        self.assertTrue(np.all(np.isclose(evol[:, 0], 0.5 * np.cos(2 * np.pi * t))))

    def test_fast_evolve(self):
        ssys = MuonSpinSystem(["mu", "e"], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        H = ssys.hamiltonian
        t = np.linspace(0, 1, 100)

        self.assertTrue(isinstance(H, CelioHamiltonian))

        # Start along z
        evol = H.fast_evolve(ssys.sigma_mu([0, 0, 1]), t, 10)

        # This test is subject to randomness, but np.isclose appears to avoid
        # any issues
        self.assertTrue(np.all(np.isclose(evol[:], 0.5 * np.cos(2 * np.pi * t))))

    def test_integrate(self):
        ssys = SpinSystem(["e"], celio_k=10)
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        H = ssys.hamiltonian
        rho0 = DensityOperator.from_vectors()  # Start along z

        with self.assertRaises(NotImplementedError):
            H.integrate_decaying(rho0, 1.0, ssys.operator({0: "z"}))



import unittest

import numpy as np
from muspinsim.spinop import DensityOperator, SpinOperator

from muspinsim.validation import (
    validate_evolve_params,
    validate_integrate_decaying_params,
    validate_times,
)


class TestValidation(unittest.TestCase):
    def test_validate_times(self):
        # Should work
        validate_times(np.array([0, 1]))

        # Invalid
        with self.assertRaises(ValueError):
            validate_times(np.eye(2))
        with self.assertRaises(TypeError):
            validate_times(
                [10],
            )

    def test_validate_evolve_params(self):
        # Should work
        validate_evolve_params(
            DensityOperator(np.eye(2)), np.array([0, 1]), [SpinOperator.from_axes()]
        )

        # Invalid rho0
        with self.assertRaises(TypeError):
            validate_evolve_params(10, np.array([0, 1]), [SpinOperator.from_axes()])

        # Invalid times
        with self.assertRaises(ValueError):
            validate_evolve_params(
                DensityOperator(np.eye(2)), np.eye(2), [SpinOperator.from_axes()]
            )

        # Invalid SpinOperators
        with self.assertRaises(ValueError):
            validate_evolve_params(
                DensityOperator(np.eye(2)),
                np.array([0, 1]),
                [SpinOperator.from_axes(), 2],
            )

    def test_validate_integrate_decaying_params(self):
        # Should work
        validate_integrate_decaying_params(
            DensityOperator(np.eye(2)), 10, [SpinOperator.from_axes()]
        )

        # Invalid rho0
        with self.assertRaises(TypeError):
            validate_integrate_decaying_params(10, 10, [SpinOperator.from_axes()])

        # Invalid tau
        with self.assertRaises(ValueError):
            validate_integrate_decaying_params(
                DensityOperator(np.eye(2)), "test", [SpinOperator.from_axes()]
            )

        # Invalid SpinOperators
        with self.assertRaises(ValueError):
            validate_integrate_decaying_params(
                DensityOperator(np.eye(2)),
                10,
                [SpinOperator.from_axes(), 2],
            )



import unittest
import numpy as np

from muspinsim.spinop import SpinOperator, DensityOperator, Operator, SuperOperator


class TestSpinOperator(unittest.TestCase):
    def test_construct(self):

        data = np.eye(4)

        s = SpinOperator(data)
        self.assertEqual(s.dimension, (4,))

        s = SpinOperator(data, dim=(2, 2))
        self.assertEqual(s.dimension, (2, 2))
        self.assertEqual(s.N, 4)

        with self.assertRaises(ValueError):
            SpinOperator(np.array([[1, 2]]))  # Not square

        with self.assertRaises(ValueError):
            SpinOperator(data, dim=(2, 3))  # Incompatible dimensions

    def test_clone(self):

        s = SpinOperator.from_axes()

        s2 = s.clone()

        self.assertIsInstance(s2, SpinOperator)
        self.assertIsInstance(s2, Operator)
        self.assertNotIsInstance(s2, DensityOperator)

    def test_herm(self):

        data = np.array([[1, 1.0j], [0, 1]])

        s = SpinOperator(data)
        self.assertFalse(s.is_hermitian)
        s = SpinOperator(data + data.T.conj())
        self.assertTrue(s.is_hermitian)

    def test_pauli(self):

        sx = SpinOperator.from_axes(0.5, "x")
        sy = SpinOperator.from_axes(0.5, "y")
        sz = SpinOperator.from_axes(0.5, "z")

        self.assertTrue(np.all(sx.matrix.toarray() == [[0, 0.5], [0.5, 0]]))
        self.assertTrue(np.all(sy.matrix.toarray() == [[0, -0.5j], [0.5j, 0]]))
        self.assertTrue(np.all(sz.matrix.toarray() == [[0.5, 0], [0, -0.5]]))

    def test_spin1(self):

        Sx = SpinOperator.from_axes(1, "x")
        Sz = SpinOperator.from_axes(1, "z")

        self.assertTrue(np.all(Sz.matrix.toarray() == np.diag([1, 0, -1])))
        self.assertTrue(
            np.all(
                np.isclose(
                    Sx.matrix.toarray(),
                    np.array([[0, 1, 0.0], [1, 0, 1.0], [0, 1, 0.0]]) / 2**0.5,
                )
            )
        )

    def test_operations(self):

        sx = SpinOperator.from_axes(0.5, "x")
        sy = SpinOperator.from_axes(0.5, "y")
        sz = SpinOperator.from_axes(0.5, "z")

        # Scalar operations
        self.assertTrue(np.all((2 * sx).matrix.toarray() == [[0, 1], [1, 0]]))
        self.assertTrue(np.all((sx / 2).matrix.toarray() == [[0, 0.25], [0.25, 0]]))

        # Operators (test commutation relations)
        self.assertTrue(
            np.all((sx * sy - sy * sx).matrix.toarray() == (1.0j * sz).matrix)
        )
        self.assertTrue(
            np.all((sy * sz - sz * sy).matrix.toarray() == (1.0j * sx).matrix)
        )
        self.assertTrue(
            np.all((sz * sx - sx * sz).matrix.toarray() == (1.0j * sy).matrix)
        )

        self.assertTrue(np.all((sx + 0.5).matrix.toarray() == 0.5 * np.ones((2, 2))))
        self.assertTrue(np.all((sz - 0.5).matrix.toarray() == np.diag([0, -1])))

        # Test equality
        self.assertTrue(
            np.allclose(sx.matrix.data, SpinOperator.from_axes(0.5, "x").matrix.data)
        )
        self.assertFalse(SpinOperator(np.eye(4)) == SpinOperator(np.eye(4), (2, 2)))

        # Test Kronecker product
        sxsz = sx.kron(sz)
        self.assertEqual(sxsz.dimension, (2, 2))
        self.assertTrue(
            np.all(
                4 * sxsz.matrix.toarray()
                == [[0, 0, 1, 0], [0, 0, 0, -1], [1, 0, 0, 0], [0, -1, 0, 0]]
            )
        )

        # Test Hilbert-Schmidt product
        rho = DensityOperator.from_vectors(0.5, np.array([1, 1, 0]) / 2**0.5)
        sx = SpinOperator.from_axes(0.5, "x")
        self.assertEqual(2 * np.real(rho.hilbert_schmidt(sx)), 0.5**0.5)

    def test_multi(self):

        Sx = SpinOperator.from_axes()
        Ix = SpinOperator.from_axes(1.0, "x")
        SxIx = SpinOperator.from_axes([0.5, 1.0], "xx")
        SxSx = SpinOperator.from_axes([0.5, 0.5], "xx")
        SySy = SpinOperator.from_axes([0.5, 0.5], "yy")

        self.assertTrue(np.allclose(Sx.kron(Ix).matrix.data, SxIx.matrix.data))

        with self.assertRaises(ArithmeticError):
            _ = Sx * SxIx

        self.assertTrue(
            np.allclose((16 * SxSx * SySy).matrix.toarray(), np.diag([-1, 1, 1, -1]))
        )

    def test_density(self):

        rho = DensityOperator(np.eye(6) / 6.0, (2, 3))
        rhosmall = rho.partial_trace([1])

        self.assertEqual(rhosmall.dimension, (2,))
        self.assertTrue(np.all(np.isclose(rhosmall.matrix.toarray(), np.eye(2) / 2)))

        with self.assertRaises(ValueError):
            DensityOperator(np.array([[0, 1], [1, 0]]))

        with self.assertRaises(ValueError):
            DensityOperator(np.array([[1, 1], [0, 1]]))

        rho = DensityOperator.from_vectors(0.5, [1, 0, 0])

        self.assertTrue(np.all(rho.matrix.toarray() == np.ones((2, 2)) * 0.5))

        rho = DensityOperator.from_vectors(0.5, [0, 1, 0], 0.5)

        self.assertTrue(
            np.all(
                np.isclose(
                    rho.matrix.toarray(), np.array([[0.5, -0.25j], [0.25j, 0.5]])
                )
            )
        )

    def test_superoperator(self):

        sx = SpinOperator.from_axes()
        rho0 = DensityOperator.from_vectors()
        lsx = SuperOperator.left_multiplier(sx)
        rsx = SuperOperator.right_multiplier(sx)
        csx = SuperOperator.commutator(sx)
        acsx = SuperOperator.anticommutator(sx)
        bksx = SuperOperator.bracket(sx)

        self.assertTrue(np.allclose((sx * rho0).matrix.data, (lsx * rho0).matrix.data))
        self.assertTrue(
            np.allclose((sx * rho0 - rho0 * sx).matrix.data, (csx * rho0).matrix.data)
        )
        self.assertTrue(
            np.allclose((sx * rho0 + rho0 * sx).matrix.data, (acsx * rho0).matrix.data)
        )
        self.assertTrue(
            np.allclose(
                (sx * rho0 * sx.dagger()).matrix.data, (bksx * rho0).matrix.data
            )
        )



from dataclasses import dataclass
import re
from typing import IO, List, Optional, Union
from numpy.typing import ArrayLike
import numpy as np


@dataclass
class GIPAWAtom:
    """
    Stores data for an atom from the output
    """

    index: int  # Starts from 1
    efg_tensor: ArrayLike


class GIPAWOutput:
    """
    Parses the output of a GIPAW execution. Currently made specifically
    for parsing the results of an EFG calculation. No guarantees this will
    work.
    """

    _atoms: List[GIPAWAtom]

    def __init__(self, file_io: Union[str, IO]):
        """Load a file containing output from GIPAW

        Reads in EFG tensors from the file.

        Arguments:
            file_io {TextIOBase} -- File path, or IO stream
        """

        self._atoms = []

        if isinstance(file_io, str):
            file_io = open(file_io, encoding="utf_8")

        # Attempt to skip to relevant part of the file
        while not file_io.readline().strip().startswith("----- total EFG -----"):
            continue

        loop = True
        while loop:
            # Should be 3 lines at a time per element e.g.
            #      V    1       -0.008879        0.000000       -0.011803
            #      V    1        0.000000       -0.030597        0.000000
            #      V    1       -0.011803        0.000000        0.039476

            lines = [
                file_io.readline().strip(),
                file_io.readline().strip(),
                file_io.readline().strip(),
            ]

            atom = self._parse_efg(lines)
            if atom is not None:
                self._atoms.append(atom)
            else:
                loop = False

            # Go to the next line
            file_io.readline()

        file_io.close()

    def _parse_efg(self, lines: List[str]) -> Optional[GIPAWAtom]:
        """Parse an EFG tensor given the 3 lines that represent it e.g.

        V    1       -0.008879        0.000000       -0.011803
        V    1        0.000000       -0.030597        0.000000
        V    1       -0.011803        0.000000        0.039476

        Arguments:
            lines {List[str]} -- Lines representing the above.
        Returns:
            GIPAW_EFG | None -- Structure containing the EFG tensor or None
                                in the case parsing failed and its likely we
                                are at the end of that section.
        Raises:
            ValueError -- When the lines have different indices or symbols.
                          Modifies indices to start from 0 instead of 1.
        """

        identifier = None
        index = None
        efg_tensor = np.zeros((3, 3))

        for i in range(0, 3):
            # Split by at least one space
            split = re.split(r"\s{1,}", lines[i])

            # Return None when invalid i.e. likely finished the section describing
            # the EFG tensors
            if len(split) != 5:
                return None

            current_identifier = split[0]
            try:
                current_index = int(split[1])
            except ValueError:
                return None

            if identifier is None:
                identifier = current_identifier
                index = current_index
            elif current_identifier != identifier or current_index != index:
                raise ValueError("Error while parsing EFG")

            efg_tensor[i, 0] = float(split[2])
            efg_tensor[i, 1] = float(split[3])
            efg_tensor[i, 2] = float(split[4])

        # Start indices from 0
        return GIPAWAtom(index=index, efg_tensor=efg_tensor)

    def find_atom(self, index: int) -> Optional[GIPAWAtom]:
        """Attempt to find an EFG tensor for atom with the given index.

        First assumes the relevant element is at the given index,
        otherwise searches for it.

        Arguments:
            index {int} -- Index of the relevant atom (starting from 1)
        Returns:
            ArrayLike | None -- EFG tensor or None if not found
        """
        # Check if at index (faster)
        if index < len(self._atoms) and self._atoms[index - 1].index == index:
            return self._atoms[index - 1]

        found_atoms = list(
            filter(lambda atom_efg: atom_efg.index == index, self._atoms)
        )
        if found_atoms:
            return found_atoms[0]
        return None



import unittest
import numpy as np
from io import StringIO
from tempfile import NamedTemporaryFile

from muspinsim.input.keyword import (
    MuSpinKeyword,
    MuSpinEvaluateKeyword,
    MuSpinExpandKeyword,
    MuSpinCouplingKeyword,
    InputKeywords,
)
from muspinsim.input.input import MuSpinInputError
from muspinsim.input import MuSpinInput


class TestInput(unittest.TestCase):
    def test_basic_keyword(self):
        # Basic keyword
        kw = MuSpinKeyword(["a b c", "d e f"])

        self.assertTrue((kw.evaluate() == [["a", "b", "c"], ["d", "e", "f"]]).all())
        self.assertEqual(len(kw), 2)

    def test_keyword_defaults(self):
        # Test that the default works
        class DefKeyword(MuSpinKeyword):
            default = "1"

        self.assertEqual(DefKeyword().evaluate()[0][0], "1")

    def test_no_defaults(self):
        with self.assertRaises(RuntimeError) as err:
            MuSpinKeyword()
        self.assertEqual(
            str(err.exception),
            "Input is empty and keyword 'keyword' doesn't have a default value",
        )

    def test_evaluate_keyword(self):
        nkw = MuSpinEvaluateKeyword(["exp(0) 1+1 2^2"])
        self.assertTrue((nkw.evaluate()[0] == [1, 2, 4]).all())

    def test_expand_keyword(self):
        exkw = MuSpinExpandKeyword(["range(0, 1)", "5.0 2.0"])
        self.assertTrue(len(exkw.evaluate()) == 101)

        def _repeat3(x):
            return [x, x, x]

        class RepeatKW(MuSpinExpandKeyword):
            _functions = {"repeat3": _repeat3}

        rkw = RepeatKW(["repeat3(1)"])
        self.assertTrue((rkw.evaluate()[0] == [1, 1, 1]).all())

    def test_invalid_number_of_args(self):
        with self.assertRaises(RuntimeError) as err:
            MuSpinKeyword([], args=["a"])  # One argument too much
        self.assertEqual(
            str(err.exception),
            "Wrong number of in-line arguments given 'a', expected 0, got 1",
        )

    def test_invalid_args(self):
        with self.assertRaises(RuntimeError) as err:
            MuSpinCouplingKeyword([], args=["a"])  # Wrong argument type
        self.assertEqual(
            str(err.exception),
            "Error parsing keyword argument(s) 'coupling_keyword': invalid literal for "
            "int() with base 10: 'a'",
        )

    def _eval_kw(self, test):
        kw = InputKeywords[test["kw"]](test["in"], args=test["args"])
        self.assertTrue((kw.evaluate() == test["out"]).all())

        self.assertEqual(
            kw.id, test["kw"] + "".join(["_{0}".format(i) for i in test["args"]])
        )

    def test_keyword_name(self):
        self._eval_kw(
            {
                "kw": "name",
                "args": [],
                "in": ["othername"],
                "out": "othername",
            }
        )

    def test_keyword_name_defaults(self):
        self._eval_kw(
            {
                "kw": "name",
                "args": [],
                "in": [],
                "out": "muspinsim",
            }
        )

    def test_keyword_spins(self):
        self._eval_kw(
            {
                "kw": "spins",
                "args": [],
                "in": ["F mu F"],
                "out": np.array(["F", "mu", "F"]),
            }
        )

    def test_keyword_spins_defaults(self):
        self._eval_kw(
            {
                "kw": "spins",
                "args": [],
                "in": [],
                "out": np.array(["mu", "e"]),
            }
        )

    def test_keyword_polarization(self):
        self._eval_kw(
            {
                "kw": "polarization",
                "args": [],
                "in": ["0 1 1"],
                "out": np.array([0, 1, 1]),
            }
        )

    def test_keyword_polarization_defaults(self):
        self._eval_kw(
            {
                "kw": "polarization",
                "args": [],
                "in": [],
                "out": np.array([1, 0, 0]),
            }
        )

    def test_keyword_polarization_longditudinal(self):
        self._eval_kw(
            {
                "kw": "polarization",
                "args": [],
                "in": ["longitudinal"],
                "out": np.array([0, 0, 1.0]),
            }
        )

    def test_keyword_polarization_transverse(self):
        self._eval_kw(
            {
                "kw": "polarization",
                "args": [],
                "in": ["transverse"],
                "out": np.array([1.0, 0, 0]),
            }
        )

    def test_keyword_field_range(self):
        self._eval_kw(
            {
                "kw": "field",
                "args": [],
                "in": ["range(0, 20, 21)"],
                "out": np.arange(21)[:, None],
            }
        )

    def test_keyword_field_mhz(self):
        kw = InputKeywords["field"](["500*MHz"])
        self.assertTrue(np.isclose(kw.evaluate()[0][0], 1.84449016))

    def test_keyword_field_defaults(self):
        self._eval_kw(
            {
                "kw": "field",
                "args": [],
                "in": [],
                "out": np.array([0]),
            }
        )

    def test_keyword_time_defaults(self):
        self._eval_kw(
            {
                "kw": "time",
                "args": [],
                "in": [],
                "out": np.linspace(0, 10, 101)[:, None],
            }
        )

    def test_keyword_time_range(self):
        self._eval_kw(
            {
                "kw": "time",
                "args": [],
                "in": ["range(0, 10, 5)"],
                "out": np.array([0, 2.5, 5, 7.5, 10])[:, None],
            }
        )

    def test_keyword_time_multi_line(self):
        self._eval_kw(
            {
                "kw": "time",
                "args": [],
                "in": ["10", "20", "30", "range(0, 10, 5)"],
                "out": np.array([10, 20, 30, 0, 2.5, 5, 7.5, 10])[:, None],
            }
        )

    def test_keyword_y_axis(self):
        self._eval_kw(
            {
                "kw": "y_axis",
                "args": [],
                "in": ["asymmetry"],
                "out": np.array(["asymmetry"]),
            }
        )

    def test_keyword_y_axis_invalid(self):
        with self.assertRaises(ValueError) as err:
            InputKeywords["y_axis"](
                ["something"], args=[]
            )  # Invalid value for argument
        self.assertEqual(
            str(err.exception),
            "Invalid value '['something']', accepts ['asymmetry', 'integral']",
        )

    def test_keyword_y_axis_defaults(self):
        self._eval_kw(
            {
                "kw": "y_axis",
                "args": [],
                "in": [],
                "out": np.array(["asymmetry"]),
            }
        )

    def test_keyword_orientation_zcw(self):
        kw = InputKeywords["orientation"](["zcw(20)"])
        self.assertTrue(len(kw.evaluate()) >= 20)

    def test_keyword_orientation_defaults(self):
        self._eval_kw(
            {
                "kw": "orientation",
                "args": [],
                "in": [],
                "out": np.array([0, 0, 0]),
            }
        )

    def test_keyword_zeeman(self):
        self._eval_kw(
            {
                "kw": "zeeman",
                "args": ["1"],
                "in": ["0 0 1"],
                "out": np.array([0, 0, 1]),
            }
        )

    def test_keyword_zeeman_defaults(self):
        self._eval_kw(
            {
                "kw": "zeeman",
                "args": ["1"],
                "in": [],
                "out": np.array([0, 0, 0]),
            }
        )

    def test_keyword_hyperfine(self):
        self._eval_kw(
            {
                "kw": "hyperfine",
                "args": ["1"],
                "in": ["1 0 0", "0 1 0", "0 0 1"],
                "out": np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
            }
        )

    def test_keyword_hyperfine_defaults(self):
        self._eval_kw(
            {
                "kw": "hyperfine",
                "args": ["1"],
                "in": [],
                "out": np.zeros((3, 3)),
            }
        )

    def test_keyword_dipolar(self):
        self._eval_kw(
            {
                "kw": "dipolar",
                "args": ["1", "2"],
                "in": ["0 0 1"],
                "out": np.array([0, 0, 1]),
            }
        )

    def test_keyword_dipolar_defaults(self):
        self._eval_kw(
            {
                "kw": "dipolar",
                "args": ["1", "2"],
                "in": [],
                "out": np.array([0, 0, 0]),
            }
        )

    def test_keyword_quadrupolar(self):
        kw = InputKeywords["quadrupolar"](
            ["1 0 0", "0 1 0", "0 0 cos(1)^2+sin(1)^2"], args=["1"]
        )
        self.assertTrue(np.allclose(kw.evaluate()[0], np.eye(3)))
        self.assertTrue(kw.id, "quadrupolar_1")

    def test_keyword_quadrupolar_defaults(self):

        self._eval_kw(
            {
                "kw": "zeeman",
                "args": ["1"],
                "in": [],
                "out": np.array([0, 0, 0]),
            }
        )

    def test_input_valid(self):
        # read valid file
        e1 = MuSpinInput(
            StringIO(
                """
name
    test_1
spins
    mu H
zeeman 1
    1 0 0
"""
            )
        ).evaluate()

        self.assertEqual(e1["name"].value[0], "test_1")
        self.assertTrue((e1["spins"].value[0] == ["mu", "H"]).all())
        self.assertTrue((e1["couplings"]["zeeman_1"].value[0] == [1, 0, 0]).all())

    def test_input_invalid_formatting(self):
        # read improperly formatted file
        with self.assertRaises(RuntimeError) as err:
            MuSpinInput(
                StringIO(
                    """
    name
        test_1
spins
    mu H
zeeman 1
    1 0 0
"""
                )
            ).evaluate()
        self.assertEqual(str(err.exception), "Badly formatted input file")

    def test_input_invalid_indent_inside_block(self):
        # indent in between keyword values does not match
        with self.assertRaises(RuntimeError) as err:
            MuSpinInput(
                StringIO(
                    """
name
    test_1
spins
    mu H
hyperfine 1
    1 0 0
     2 0 0
    3 0 0
"""
                )
            ).evaluate()
        self.assertEqual(str(err.exception), "Invalid indent in input file")

    def test_input_invalid_keyword(self):
        # incorrect keyword name given
        with self.assertRaises(MuSpinInputError) as err:
            MuSpinInput(
                StringIO(
                    """
name
    test_1
spins
    mu H
notakeyword 1
    1 0 0
"""
                )
            ).evaluate()
        self.assertEqual(
            str(err.exception),
            "Found 1 error(s) whilst trying to parse keywords: \n\n"
            "Error occurred when parsing keyword 'notakeyword' "
            "(block starting at line 6):\n"
            "Invalid keyword 'notakeyword' found in input file",
        )

    def test_input_fitting(self):
        # Test input focused around fitting

        i1 = MuSpinInput(
            StringIO(
                """
fitting_variables
    x 1.0 0.0 2.0
fitting_data
    0  0.0
    1  1.0
    2  4.0
    3  9.0
field
    2*x
zeeman 1
    x x 0
"""
            )
        )

        self.assertTrue(i1.fitting_info["fit"])

        data = i1.fitting_info["data"]
        self.assertTrue((data == [[0, 0], [1, 1], [2, 4], [3, 9]]).all())

        e1 = i1.evaluate(x=2.0)
        self.assertEqual(e1["field"].value[0][0], 4.0)
        self.assertTrue((e1["couplings"]["zeeman_1"].value[0] == [2, 2, 0]).all())

        variables = i1.variables

        self.assertEqual(variables["x"].value, 1.0)
        self.assertEqual(variables["x"].bounds, (0.0, 2.0))

    def _write_temp_file(self, tdata):
        tfile = NamedTemporaryFile(mode="w", delete=False)

        for d in tdata:
            tfile.write("{0} {1}\n".format(*d))
        tfile.flush()
        tfile.close()
        return tfile

    def test_load_fitting_data_from_file(self):
        tdata = np.zeros((10, 2))
        tdata[:, 0] = np.linspace(0, 1, 10)
        tdata[:, 1] = tdata[:, 0] ** 2

        tfile = self._write_temp_file(tdata)

        i2 = MuSpinInput(
            StringIO(
                """
fitting_variables
    x
fitting_data
    load("{fname}")
fitting_method
    nelder-mead
""".format(
                    fname=tfile.name
                )
            )
        )

        finfo = i2.fitting_info

        data = finfo["data"]
        self.assertTrue(finfo["fit"])
        self.assertTrue((data == tdata).all())
        self.assertEqual(finfo["method"], "nelder-mead")
        self.assertAlmostEqual(finfo["rtol"], 1e-3)

    def test_fitting_no_data(self):
        # invalid no data given
        with self.assertRaises(MuSpinInputError) as err:
            MuSpinInput(
                StringIO(
                    """
fitting_variables
    x 1.0 0.0 5.0
"""
                )
            )
        self.assertEqual(
            str(err.exception),
            "Found 1 Error(s) whilst trying to parse fitting keywords: \n\n"
            "Error occurred when parsing keyword 'fitting_variables' "
            "(block starting at line 2):\n"
            "Fitting variables defined without defining any data to fit",
        )

    def test_fitting_invalid_variable_ranges(self):
        tdata = np.zeros((10, 2))
        tdata[:, 0] = np.linspace(0, 1, 10)
        tdata[:, 1] = tdata[:, 0] ** 2
        tfile = self._write_temp_file(tdata)

        with self.assertRaises(MuSpinInputError) as err:
            MuSpinInput(
                StringIO(
                    """
fitting_variables
    x 1.0 0.0 -5.0
fitting_data
    load("{fname}")
fitting_method
    nelder-mead
field
    2*x
zeeman 1
    x x 0
""".format(
                        fname=tfile.name
                    )
                )
            )
        self.assertTrue(
            "Variable x has invalid range: "
            "(max value -5.0 cannot be less than or equal to min value 0.0)\n"
            "Variable x has invalid starting value: "
            "(starting value 1.0 cannot be greater than max value -5.0)"
            in str(err.exception)
        )

    def test_fitting_name_clash(self):
        # variable name clashes with constant
        with self.assertRaises(MuSpinInputError) as err:
            MuSpinInput(
                StringIO(
                    """
fitting_variables
    MHz 1.0 0.0 5.0
fitting_data
    0  0.0
    1  1.0
    2  4.0
    3  9.0
field
    2*x
zeeman 1
    MHz 0 0
"""
                )
            )
        self.assertEqual(
            str(err.exception),
            "Found 1 Error(s) whilst trying to parse fitting keywords: \n\n"
            "Error occurred when parsing keyword 'fitting_variables' "
            "(block starting at line 2):\n"
            "Invalid value 'MHz': variable name conflicts with a constant",
        )



"""mpi.py

Functions useful for OPEN MPI parallelisation"""

import warnings
import numpy as np


class MPIController:
    def __init__(self):
        self._MPI = None
        self._comm = None
        self._rank = 0
        self._size = 1

    def connect(self):
        try:
            from mpi4py import MPI
        except ImportError:
            warnings.warn("Can not use MPI; please install mpi4py")
            return

        self._MPI = MPI
        self._comm = MPI.COMM_WORLD
        self._rank = self.comm.Get_rank()
        self._size = self.comm.Get_size()

    @property
    def comm(self):
        return self._comm

    @property
    def rank(self):
        return self._rank

    @property
    def size(self):
        return self._size

    @property
    def is_root(self):
        return self._rank == 0

    def execute_on_root(self, func):
        """A decorator making sure a function only executes on the root node"""

        def decfunc(*args, **kwargs):
            if not self.is_root:
                return None
            return func(*args, **kwargs)

        return decfunc

    def broadcast(self, var):
        """A function to broadcast a single variable"""

        if self.comm is not None:
            var = self.comm.bcast(var, root=0)

        return var

    def broadcast_object(self, obj, selected_attrs=None):
        """A function to broadcast an object's members"""

        if self.comm is None:
            return  # Nothing to do

        if selected_attrs is None:
            selected_attrs = list(obj.__dict__.keys())

        selected_attrs = self.comm.bcast(selected_attrs, root=0)

        for k in selected_attrs:
            v = obj.__dict__.get(k, None)
            try:
                v = self.comm.bcast(v, root=0)
            except OverflowError as exc:
                raise OverflowError(f"Overflow error on: {k}, {v}") from exc
            obj.__setattr__(k, v)

    def broadcast_terms(self, lst):
        if self.comm is None:
            return lst

        num_list = self.comm.bcast(len(lst), root=0)

        n_list = []
        for i in range(num_list):

            # broadcast object attributes
            if self.is_root:
                # spinsys cannot be broadcast - issue with recursive object references?
                lst[i].__setattr__("_spinsys", None)
                val = lst[i]
            else:
                val = None

            try:
                val = self.comm.bcast(val, root=0)
            except OverflowError as exc:
                raise OverflowError(f"Overflow error broadcasting term {val}") from exc
            n_list.append(val)
        return n_list

    def sum_data(self, data):

        if self._MPI is None:
            return data

        # Otherwise, create a dummy copy
        data_dest = data.copy()
        self.comm.Reduce(data, data_dest, op=self._MPI.SUM, root=0)

        return data_dest

    def split_1D(self, vector, size=None):

        n = size if size else self.size

        if n == 1:
            return [vector]

        a = len(vector)
        q = int(np.floor(a / n))
        r = a % n

        sizes = np.array([q] * n)
        sizes[:r] += 1
        ends = np.cumsum(sizes)

        i = 0
        s_i = 0
        split = []
        while s_i < n:
            split.append(vector[slice(i, ends[s_i])])
            i = ends[s_i]
            s_i += 1

        return split

    def split_2D(self, vector1, vector2, size=None):

        n = size if size else self.size

        if n == 1:
            return [(vector1, vector2)]

        # Now let's go for the actual parallelism
        a = len(vector1)
        b = len(vector2)
        # Optimal number of operations per core?
        c = int(np.ceil(a * b / n))

        candidates = []

        for s1 in range(1, n + 1):
            s2 = n / s1
            if s2 != int(s2):
                continue
            s2 = int(s2)

            split1 = self.split_1D(vector1, s1)
            split2 = self.split_1D(vector2, s2)

            sizes1 = [len(s) for s in split1]
            sizes2 = [len(s) for s in split2]

            areas = np.array([s1 * s2 for s1 in sizes1 for s2 in sizes2])

            # Score based on how far they go from the optimal 'area' c
            candidates.append((split1, split2, np.sum((areas - c) ** 2)))

        candidates = sorted(candidates, key=lambda x: x[2])

        split1, split2 = candidates[0][:2]

        split = []
        for sa in split1:
            for sb in split2:
                split.append((sa, sb))

        return split


mpi_controller = MPIController()



from abc import ABC, abstractmethod
import argparse
from dataclasses import dataclass, field
import logging
import sys
from typing import List

from muspinsim.input.structure import CellAtom, MuonatedStructure
from muspinsim.input.gipaw import GIPAWOutput


class InteractionGenerator(ABC):
    """
    Abstract base class for handling the generation of muspinsim interaction
    config
    """

    @abstractmethod
    def gen_config(
        self, muon_file_index: int, atom_file_index: int, atom: CellAtom
    ) -> str:
        """Should return the config required for an interaction between the muon
        and given atom using this generator

        Arguments:
            muon_file_index {int} -- Index of the muon in the file (starts
                                     from 1)
            atom_file_index {int} -- Index of the atom in the file (starts
                                     from 1)
            atom {CellAtom} -- Atom to generate config for
        """


class DipoleIntGenerator(InteractionGenerator):
    """
    Generator for the dipole interaction
    """

    def gen_config(
        self, muon_file_index: int, atom_file_index: int, atom: CellAtom
    ) -> str:
        return f"""dipolar {muon_file_index} {atom_file_index}
    {" ".join(map(str, atom.vector_from_muon))}"""


class QuadrupoleIntGenerator(InteractionGenerator):
    """
    Generator for the quadrupole interaction
    """

    _gipaw_file: GIPAWOutput

    def __init__(self, gipaw_file: GIPAWOutput):
        self._gipaw_file = gipaw_file

    def gen_config(
        self, muon_file_index: int, atom_file_index: int, atom: CellAtom
    ) -> str:
        # Obtain corresponding EFG tensor
        gipaw_atom = self._gipaw_file.find_atom(atom.index)

        if gipaw_atom is None:
            raise ValueError(
                f"Unable to locate atom with index {atom.index} in GIPAW output file"
            )
        efg_tensor = gipaw_atom.efg_tensor

        return f"""quadrupolar {atom_file_index}
    {' '.join(map(str, efg_tensor[0]))}
    {' '.join(map(str, efg_tensor[1]))}
    {' '.join(map(str, efg_tensor[2]))}"""


@dataclass
class GeneratorToolParams:
    """Structure for storing parameters for the interaction generator tool

    Arguments:
        structure {MuonatedStructure} -- Structure to compute terms from
        generators {List[InteractionGenerator]} -- List of generators for
                                                generating interaction terms
        number_closest {int} -- Number of closest atoms to the muon to include
        additional_ignored_symbols {List[str]} -- List of additional symbols
                                    to ignore when counting the closest atoms.
                                    All spin 0 isotopes will be ignored by
                                    default but currently this data is not
                                    reflected by soprano for Si.
        max_layer {int} -- Maximum layer to allow the expansion to (for
                           avoiding any long compute times)
    """

    structure: MuonatedStructure
    generators: List[InteractionGenerator]
    number_closest: int
    additional_ignored_symbols: List[str] = field(default_factory=list)
    max_layer: int = 6


def generate_input_file(params: GeneratorToolParams) -> str:
    """Utility function for generating muspinsim input config given a
    structure as input

    Will expand the muonated structure by the amount requested and
    locate the nearest neighbours in order to generate their interactions
    for the config file.

    Arguments:
        params {GeneratorToolParams} -- Parameters (see above for details)
    """

    # Locate closest elements (ignoring ones with zero spin)
    selected_atoms = params.structure.compute_closest(
        number=params.number_closest,
        ignored_symbols=params.structure.symbols_zero_spin
        + params.additional_ignored_symbols,
        max_layer=params.max_layer,
    )

    # Generate input file
    selected_symbols = "mu"
    muon_file_index = 1
    atom_file_index = 2

    input_file = ""

    for selected_atom in selected_atoms:
        symbol = selected_atom.symbol

        if selected_atom.isotope != 1:
            symbol = f"{str(selected_atom.isotope)} {symbol}"

        selected_symbols += f" {symbol}"

        for generator in params.generators:
            input_file += f"""{
                generator.gen_config(
                    muon_file_index,
                    atom_file_index,
                    selected_atom
                )
            }\n"""

        atom_file_index += 1

    input_file = f"""spins
    {selected_symbols}
{input_file}"""

    return input_file


def _run_generator_tool(args):
    # Setup so we can see the log output
    logging.basicConfig(
        format="[%(levelname)s] [%(asctime)s] %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    parser = argparse.ArgumentParser(
        description="""Generate a MuSpinSim input file using a structure file
(.cell or .cif).""",
    )

    # Required arguments
    parser.add_argument("filepath", type=str, help="Structure file filepath")
    parser.add_argument(
        "number_closest",
        type=int,
        help="Number of nearest atoms to the muon stopping site to include",
    )
    # Optional arguments
    parser.add_argument(
        "--dipolar",
        dest="dipolar",
        action="store_true",
        help="Specify to include dipolar couplings.",
    )
    parser.add_argument(
        "--quadrupolar",
        dest="gipaw_filepath",
        type=str,
        help="""Filepath for GIPAW output data defining the EFGs to be used to
include quadrupolar couplings.""",
    )
    parser.add_argument(
        "--out",
        dest="output",
        type=str,
        help="Filepath for the output file to generate",
    )
    parser.add_argument(
        "--muon_symbol",
        dest="muon_symbol",
        type=str,
        help="Symbol used to identify the muon (default: H)",
        default="H",
    )
    parser.add_argument(
        "--ignore_symbol",
        dest="ignored_symbols",
        action="append",
        help="""Allows certain elements in the cell file to be ignored by
their symbol when finding the closest atoms.""",
        default=[],
    )
    parser.add_argument(
        "--max_layer",
        dest="max_layer",
        type=int,
        help="""Maximum number of layers the structure can be expanded to
before raising an error (default: 6). Larger values may be needed when
searching for many atoms.""",
        default=6,
    )

    args = parser.parse_args(args)

    # Terms to generate
    generators = []

    if args.dipolar:
        generators.append(DipoleIntGenerator())
    if args.gipaw_filepath is not None:
        generators.append(QuadrupoleIntGenerator(GIPAWOutput(args.gipaw_filepath)))

    generate_params = GeneratorToolParams(
        structure=MuonatedStructure(args.filepath, muon_symbol=args.muon_symbol),
        generators=generators,
        number_closest=args.number_closest,
        additional_ignored_symbols=args.ignored_symbols,
        max_layer=args.max_layer,
    )
    file_data = generate_input_file(generate_params)

    if args.output is not None:
        with open(args.output, "w", encoding="utf-8") as file:
            file.write(file_data)
    else:
        print(file_data)


def main():
    """Entrypoint for command line tool"""

    _run_generator_tool(sys.argv[1:])



import unittest
import numpy as np

from muspinsim.mpi import mpi_controller


class TestMuSpinMPI(unittest.TestCase):
    def test_split(self):
        # Test efficient splitting of processes
        a = list(range(20))
        b = list(range(17))

        split7 = mpi_controller.split_1D(a, 7)

        self.assertTrue(np.all(split7[0] == [0, 1, 2]))
        self.assertTrue(
            np.all([len(s) for s in split7] == np.array([3, 3, 3, 3, 3, 3, 2]))
        )

        split8 = mpi_controller.split_1D(a, 8)

        self.assertTrue(np.all(split8[0] == [0, 1, 2]))
        self.assertTrue(
            np.all([len(s) for s in split8] == np.array([3, 3, 3, 3, 2, 2, 2, 2]))
        )

        split30 = mpi_controller.split_2D(a, b, 30)

        self.assertTrue(np.all(split30[0][0] == [0, 1]))
        self.assertTrue(np.all(split30[0][1] == [0, 1, 2, 3, 4, 5]))
        self.assertTrue(np.all(split30[2][0] == [0, 1]))
        self.assertTrue(np.all(split30[2][1] == [12, 13, 14, 15, 16]))

    def test_broadcast(self):
        class A:
            def __init__(self, x):
                self.x = x

        class B:
            def __init__(self, x):
                self.a = A(x)

        b = B(mpi_controller.rank)
        mpi_controller.broadcast_object(b)

        self.assertEqual(b.a.x, 0)

        d = np.array([b.a.x])
        d = mpi_controller.sum_data(d)

        if mpi_controller.is_root:
            self.assertEqual(d[0], 0)



import unittest

from muspinsim import constants


class TestConstants(unittest.TestCase):
    def test_values(self):

        self.assertAlmostEqual(constants.ELEC_GAMMA, -28024.9514242)
        self.assertAlmostEqual(constants.MU_GAMMA, 135.53880943285955)
        self.assertAlmostEqual(constants.MU_TAU, 2.19703)
        self.assertAlmostEqual(constants.EFG_2_MHZ, 0.23496477815245767)

    def test_gyromagnetic(self):

        self.assertEqual(constants.gyromagnetic_ratio("e"), constants.ELEC_GAMMA)
        self.assertEqual(constants.gyromagnetic_ratio("mu"), constants.MU_GAMMA)
        self.assertAlmostEqual(constants.gyromagnetic_ratio("H"), 42.577478518, 3)

        with self.assertRaises(ValueError) as err:
            constants.gyromagnetic_ratio("H", 4)
        self.assertEqual(str(err.exception), "Invalid isotope 4 for element H")

    def test_quadrupole(self):

        self.assertEqual(constants.quadrupole_moment("e"), 0)
        self.assertEqual(constants.quadrupole_moment("H"), 0)
        self.assertAlmostEqual(constants.quadrupole_moment("H", 2), 2.859, 2)

        with self.assertRaises(ValueError) as err:
            constants.gyromagnetic_ratio("H", 4)
        self.assertEqual(str(err.exception), "Invalid isotope 4 for element H")

    def test_spin(self):

        self.assertEqual(constants.spin("mu"), 0.5)
        self.assertEqual(constants.spin("e"), 0.5)
        self.assertEqual(constants.spin("H"), 0.5)
        self.assertEqual(constants.spin("H", 2), 1.0)

        with self.assertRaises(ValueError) as err:
            constants.spin("H", 4)
        self.assertEqual(str(err.exception), "Invalid isotope 4 for element H")

        with self.assertRaises(ValueError) as err:
            constants.spin("e", 0.5)
        self.assertEqual(str(err.exception), "Invalid multiplicity 0.5 for electron")



import numpy as np


class FittingVariable:
    """Class defining a fitting variable with a starting value and bounds"""

    def __init__(self, name, value=0.0, min_value=-np.inf, max_value=np.inf):

        self._name = name
        self._value = float(value)
        self._min = float(min_value)
        self._max = float(max_value)

        invalid = ""
        if self._max <= self._min:
            invalid += (
                f"Variable {name} has invalid range: "
                f"(max value {self._max} cannot be less than or equal to min "
                f"value {self._min}"
                ")\n"
            )
        if self._value > self._max:
            invalid += (
                f"Variable {name} has invalid starting value: "
                f"(starting value {self._value} cannot be greater than max "
                f"value {self._max})"
            )
        if self._value < self._min:
            invalid += (
                f"Variable {name} has invalid starting value: "
                f"(starting value {self._value} cannot be less than min "
                f"value {self._min})"
            )

        if invalid != "":
            raise ValueError(invalid)

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @property
    def bounds(self):
        return (self._min, self._max)
